@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2015 ECMWF.
+ * Copyright 2005-2016 ECMWF.
  *
  * This software is licensed under the terms of the Apache Licence Version 2.0
  * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
@@ -8,10 +8,6 @@
  * virtue of its status as an intergovernmental organisation nor does it submit to any jurisdiction.
  */
 
-/***************************************************************************
- *   Enrico Fucile
- *                                                                         *
- ***************************************************************************/
 #include "grib_api_internal.h"
 
 #ifdef GRIB_ON_WINDOWS
@@ -369,8 +365,188 @@ static void print_values(grib_context* c, const grib_util_grid_spec* spec, const
     }
 }
 
-#define     ISECTION_2  3000
-#define     ISECTION_4  512
+static int DBL_EQUAL(double d1, double d2, double tolerance)
+{
+    return fabs(d1-d2) < tolerance;
+}
+
+/* Returns a boolean: 1 if angle can be encoded, 0 otherwise */
+static int grib1_angle_can_be_encoded(const double angle)
+{
+    const double angle_milliDegrees = angle * 1000;
+    double rounded = (int)(angle_milliDegrees+0.5)/1000.0;
+    if (angle<0) {
+        rounded = (int)(angle_milliDegrees-0.5)/1000.0;
+    }
+    if (angle == rounded) return 1;
+    return 0; /* sub millidegree. Cannot be encoded in grib1 */
+}
+
+/* Returns a boolean: 1 if angle is too small, 0 otherwise */
+static int angle_too_small(const double angle, const double angular_precision)
+{
+    const double a = fabs(angle);
+    if (a > 0 && a < angular_precision) return 1;
+    return 0;
+}
+
+static double normalise_angle(double angle)
+{
+    while (angle<0)   angle += 360;
+    while (angle>360) angle -= 360;
+    return angle;
+}
+
+/* Check what is coded in the handle is what is requested by the spec. */
+/* Return GRIB_SUCCESS if the geometry matches, otherwise the error code */
+static int check_handle_against_spec(grib_handle* handle, const long edition, const grib_util_grid_spec* spec)
+{
+    int err = 0;
+    int check_latitudes = 1;
+    int check_longitudes = 1;
+    double angular_precision = 1.0/1000.0; /* millidegree */
+    double tolerance = angular_precision/2.0; /* half a millidegree */
+
+    if (edition == 2) {
+        angular_precision = 1.0/1000000.0; /* microdegree */
+        tolerance = angular_precision/2.0;
+    }
+
+    if (edition == 2) {
+        return GRIB_SUCCESS;  /* For now only do checks on edition 1 */
+    }
+
+    if (spec->grid_type == GRIB_UTIL_GRID_SPEC_POLAR_STEREOGRAPHIC ||
+        spec->grid_type == GRIB_UTIL_GRID_SPEC_SH)
+    {
+        return GRIB_SUCCESS;
+    }
+
+    /* Cannot check latitudes of Gaussian grids because are always sub-millidegree */
+    /* and for GRIB1 will differ from the encoded values. We accept this discrepancy! */
+    if (spec->grid_type == GRIB_UTIL_GRID_SPEC_REGULAR_GG ||
+        spec->grid_type == GRIB_UTIL_GRID_SPEC_ROTATED_GG ||
+        spec->grid_type == GRIB_UTIL_GRID_SPEC_REDUCED_GG ||
+        spec->grid_type == GRIB_UTIL_GRID_SPEC_REDUCED_ROTATED_GG)
+    {
+        if (edition == 1) {
+            check_latitudes = 0;
+        }
+    }
+
+    if (check_latitudes) {
+        double lat1, lat2;
+        const double lat1spec = normalise_angle(spec->latitudeOfFirstGridPointInDegrees);
+        const double lat2spec = normalise_angle(spec->latitudeOfLastGridPointInDegrees);
+        if ((err = grib_get_double(handle, "latitudeOfFirstGridPointInDegrees", &lat1))!=0)  return err;
+        if ((err = grib_get_double(handle, "latitudeOfLastGridPointInDegrees", &lat2))!=0)   return err;
+
+        lat1 = normalise_angle(lat1);
+        lat2 = normalise_angle(lat2);
+
+        if (angle_too_small(lat1spec, angular_precision)) {
+            fprintf(stderr, "Failed to encode latitude of first grid point %.10e: less than angular precision\n",lat1spec);
+            return GRIB_WRONG_GRID;
+        }
+        if (angle_too_small(lat2spec, angular_precision)) {
+            fprintf(stderr, "Failed to encode latitude of last grid point %.10e: less than angular precision\n", lat2spec);
+            return GRIB_WRONG_GRID;
+        }
+
+        if (!DBL_EQUAL(lat1spec, lat1, tolerance)) {
+            fprintf(stderr, "Failed to encode latitude of first grid point: spec=%.10e val=%.10e\n", lat1spec, lat1);
+            return GRIB_WRONG_GRID;
+        }
+        if (!DBL_EQUAL(lat2spec, lat2, tolerance)) {
+            fprintf(stderr, "Failed to encode latitude of last grid point: spec=%.10e val=%.10e\n", lat2spec, lat2);
+            return GRIB_WRONG_GRID;
+        }
+    }
+
+    if (check_longitudes) {
+        double lon1, lon2;
+        const double lon1spec = normalise_angle(spec->longitudeOfFirstGridPointInDegrees);
+        const double lon2spec = normalise_angle(spec->longitudeOfLastGridPointInDegrees);
+        if ((err = grib_get_double(handle, "longitudeOfFirstGridPointInDegrees", &lon1))!=0) return err;
+        if ((err = grib_get_double(handle, "longitudeOfLastGridPointInDegrees", &lon2))!=0)  return err;
+
+        lon1 = normalise_angle(lon1);
+        lon2 = normalise_angle(lon2);
+
+        if (angle_too_small(lon1spec, angular_precision)) {
+            fprintf(stderr, "Failed to encode longitude of first grid point %.10e: less than angular precision\n", lon1spec);
+            return GRIB_WRONG_GRID;
+        }
+        if (angle_too_small(lon2spec, angular_precision)) {
+            fprintf(stderr, "Failed to encode longitude of last grid point %.10e: less than angular precision\n", lon2spec);
+            return GRIB_WRONG_GRID;
+        }
+
+        if (!DBL_EQUAL(lon1spec, lon1, tolerance)) {
+            fprintf(stderr, "Failed to encode longitude of first grid point: spec=%.10e val=%.10e\n", lon1spec, lon1);
+            return GRIB_WRONG_GRID;
+        }
+        if (!DBL_EQUAL(lon2spec, lon2, tolerance)){
+            fprintf(stderr, "Failed to encode longitude of last grid point: spec=%.10e val=%.10e\n",  lon2spec, lon2);
+            return GRIB_WRONG_GRID;
+        }
+    }
+
+    if (spec->grid_type == GRIB_UTIL_GRID_SPEC_ROTATED_LL ||
+        spec->grid_type == GRIB_UTIL_GRID_SPEC_ROTATED_GG ||
+        spec->grid_type == GRIB_UTIL_GRID_SPEC_REDUCED_ROTATED_GG)
+    {
+        double latp, lonp;
+        const double latspec = normalise_angle(spec->latitudeOfSouthernPoleInDegrees);
+        const double lonspec = normalise_angle(spec->longitudeOfSouthernPoleInDegrees);
+        if ((err = grib_get_double(handle, "latitudeOfSouthernPoleInDegrees", &latp))!=0)  return err;
+        if ((err = grib_get_double(handle, "longitudeOfSouthernPoleInDegrees", &lonp))!=0)  return err;
+        latp = normalise_angle(latp);
+        lonp = normalise_angle(lonp);
+
+        if (!DBL_EQUAL(latspec, latp, tolerance)) {
+            fprintf(stderr, "Failed to encode latitude of southern pole: spec=%.10e val=%.10e\n",latspec,latp);
+            return GRIB_WRONG_GRID;
+        }
+        if (!DBL_EQUAL(lonspec, lonp, tolerance)) {
+            fprintf(stderr, "Failed to encode longitude of southern pole: spec=%.10e val=%.10e\n",lonspec,lonp);
+            return GRIB_WRONG_GRID;
+        }
+    }
+    return GRIB_SUCCESS;
+}
+
+static char* get_grid_type_name(const int spec_grid_type)
+{
+    if (spec_grid_type == GRIB_UTIL_GRID_SPEC_REGULAR_LL)
+        return "regular_ll";
+
+    if (spec_grid_type == GRIB_UTIL_GRID_SPEC_ROTATED_LL)
+        return "rotated_ll";
+
+    if (spec_grid_type == GRIB_UTIL_GRID_SPEC_REGULAR_GG)
+        return "regular_gg";
+
+    if (spec_grid_type == GRIB_UTIL_GRID_SPEC_ROTATED_GG)
+        return "rotated_gg";
+
+    if (spec_grid_type == GRIB_UTIL_GRID_SPEC_REDUCED_LL)
+        return "reduced_ll";
+
+    if (spec_grid_type == GRIB_UTIL_GRID_SPEC_POLAR_STEREOGRAPHIC)
+        return "polar_stereographic";
+
+    if (spec_grid_type == GRIB_UTIL_GRID_SPEC_REDUCED_GG)
+        return "reduced_gg";
+
+    if (spec_grid_type == GRIB_UTIL_GRID_SPEC_SH)
+        return "sh";
+
+    if (spec_grid_type == GRIB_UTIL_GRID_SPEC_REDUCED_ROTATED_GG)
+        return "reduced_rotated_gg";
+
+    return NULL;
+}
 
 grib_handle* grib_util_set_spec(grib_handle* h,
         const grib_util_grid_spec    *spec,
@@ -406,6 +582,7 @@ grib_handle* grib_util_set_spec(grib_handle* h,
     int packingTypeIsSet=0;
     int setSecondOrder=0;
     size_t slen=17;
+    int grib1_high_resolution_fix = 0; /* boolean: See GRIB-863 */
 
     static grib_util_packing_spec default_packing_spec = {0, };
     Assert(h);
@@ -589,48 +766,11 @@ grib_handle* grib_util_set_spec(grib_handle* h,
         return h;
     }
 
-    switch(spec->grid_type) {
-
-    case GRIB_UTIL_GRID_SPEC_REGULAR_LL:
-        grid_type = "regular_ll";
-        break;
-
-    case GRIB_UTIL_GRID_SPEC_ROTATED_LL:
-        grid_type = "rotated_ll";
-        break;
-
-    case GRIB_UTIL_GRID_SPEC_REGULAR_GG:
-        grid_type = "regular_gg";
-        break;
-
-    case GRIB_UTIL_GRID_SPEC_ROTATED_GG:
-        grid_type = "rotated_gg";
-        break;
-
-    case GRIB_UTIL_GRID_SPEC_REDUCED_LL:
-        grid_type = "reduced_ll";
-        break;
-
-    case GRIB_UTIL_GRID_SPEC_POLAR_STEREOGRAPHIC:
-        grid_type = "polar_stereographic";
-        break;
-
-    case GRIB_UTIL_GRID_SPEC_REDUCED_GG:
-        grid_type = "reduced_gg";
-        break;
-
-    case GRIB_UTIL_GRID_SPEC_SH:
-        grid_type = "sh";
-        break;
-
-    case GRIB_UTIL_GRID_SPEC_REDUCED_ROTATED_GG:
-        grid_type = "reduced_rotated_gg";
-        break;
-
-    default:
+    grid_type = get_grid_type_name(spec->grid_type);
+    if (grid_type == NULL) {
+        fprintf(stderr,"GRIB_UTIL_SET_SPEC: Unknown grid type: %d\n", spec->grid_type);
         *err = GRIB_NOT_IMPLEMENTED;
         return NULL;
-
     }
 
     SET_STRING_VALUE("gridType", grid_type);
@@ -644,7 +784,12 @@ grib_handle* grib_util_set_spec(grib_handle* h,
         switch (spec->grid_type) {
         case GRIB_UTIL_GRID_SPEC_REDUCED_GG:
         case GRIB_UTIL_GRID_SPEC_REDUCED_ROTATED_GG:
+            /* Choose a sample with the right Gaussian number and edition */
             sprintf(name, "%s_pl_%ld_grib%ld", grid_type,spec->N, editionNumber);
+            if (spec->pl && spec->pl_size) {
+                /* GRIB-834: pl is given so can use any of the reduced_gg_pl samples */
+                sprintf(name, "%s_pl_grib%ld", grid_type, editionNumber);
+            }
             break;
         default :
             sprintf(name, "%s_pl_grib%ld", grid_type, editionNumber);
@@ -666,7 +811,17 @@ grib_handle* grib_util_set_spec(grib_handle* h,
         COPY_SPEC_LONG  (bitmapPresent);
         if (spec->missingValue) COPY_SPEC_DOUBLE(missingValue);
 
-        SET_LONG_VALUE  ("ijDirectionIncrementGiven",    1);
+        SET_LONG_VALUE("ijDirectionIncrementGiven", 1);
+        if (editionNumber == 1) {
+            /* GRIB-863: GRIB1 cannot represent increments less than a millidegree */
+            if (!grib1_angle_can_be_encoded(spec->iDirectionIncrementInDegrees) ||
+                !grib1_angle_can_be_encoded(spec->jDirectionIncrementInDegrees))
+            {
+                grib1_high_resolution_fix = 1;
+                /* Set flag to compute the increments */
+                SET_LONG_VALUE("ijDirectionIncrementGiven", 0);
+            }
+        }
 
         /* default iScansNegatively=0 jScansPositively=0 is ok */
         COPY_SPEC_LONG(iScansNegatively);
@@ -867,9 +1022,7 @@ grib_handle* grib_util_set_spec(grib_handle* h,
         }
     }
 
-
     switch(packing_spec->accuracy) {
-
     case GRIB_UTIL_ACCURACY_SAME_BITS_PER_VALUES_AS_INPUT:
     {
         long bitsPerValue = 0;
@@ -906,7 +1059,6 @@ grib_handle* grib_util_set_spec(grib_handle* h,
             values[count++] = packing_spec->extra_settings[i];
         }
     }
-
     /* grib_write_message(h,"input.grib","w"); */
     /* grib_write_message(tmp,"geo.grib","w"); */
     if((outh = grib_util_sections_copy(h, tmp, GRIB_SECTION_PRODUCT | GRIB_SECTION_LOCAL,err)) == NULL)
@@ -917,19 +1069,8 @@ grib_handle* grib_util_set_spec(grib_handle* h,
     grib_handle_delete(tmp);
     Assert(*err == 0);
 
-    if (h->context->debug==-1)
-        print_values(h->context,spec,data_values,data_values_count,values,count);
-
-    if((*err = grib_set_values(outh,values,count)) != 0)
-    {
-        fprintf(stderr,"SET_GRID_DATA_DESCRIPTION: Cannot set values  %s\n",grib_get_error_message(*err));
-
-        for(i = 0; i < count; i++)
-            if(values[i].error)
-                fprintf(stderr," %s %s\n",values[i].name,grib_get_error_message(values[i].error));
-        goto cleanup;
-    }
-
+    /* Set "pl" array if provided (For reduced Gaussian grids) */
+    /* See GRIB-857 */
     Assert(spec->pl_size >= 0);
     if (spec->pl && spec->pl_size == 0) {
         fprintf(stderr, "pl array not NULL but pl_size == 0!\n");
@@ -949,6 +1090,19 @@ grib_handle* grib_util_set_spec(grib_handle* h,
         }
     }
 
+    if (h->context->debug==-1)
+        print_values(h->context,spec,data_values,data_values_count,values,count);
+
+    if((*err = grib_set_values(outh,values,count)) != 0)
+    {
+        fprintf(stderr,"SET_GRID_DATA_DESCRIPTION: Cannot set values  %s\n",grib_get_error_message(*err));
+
+        for(i = 0; i < count; i++)
+            if(values[i].error)
+                fprintf(stderr," %s %s\n",values[i].name,grib_get_error_message(values[i].error));
+        goto cleanup;
+    }
+
     if((*err = grib_set_double_array(outh,"values",data_values,data_values_count)) != 0)
     {
         FILE* ferror;
@@ -973,16 +1127,31 @@ grib_handle* grib_util_set_spec(grib_handle* h,
     /* grib_write_message(outh,"h.grib","w"); */
     /* if the field is empty GRIBEX is packing as simple*/
     /*
-	if (!strcmp(input_packing_type,"grid_simple_matrix")) {
-		long numberOfValues;
-		grib_get_long(outh,"numberOfValues",&numberOfValues);
-		if (numberOfValues==0)  {
-			slen=11;
-			grib_set_string(outh,"packingType","grid_simple",&slen);
-		}
-	}
+    if (!strcmp(input_packing_type,"grid_simple_matrix")) {
+        long numberOfValues;
+        grib_get_long(outh,"numberOfValues",&numberOfValues);
+        if (numberOfValues==0)  {
+            slen=11;
+            grib_set_string(outh,"packingType","grid_simple",&slen);
+        }
+    }
      */
 
+    if (grib1_high_resolution_fix) {
+        /* GRIB-863: must set increments to MISSING */
+        /* increments are not coded in message but computed */
+        if ( (*err = grib_set_missing(outh, "iDirectionIncrement"))!=0 ) {
+            fprintf(stderr,"GRIB_UTIL_SET_SPEC: Cannot set Di to missing: %s\n",grib_get_error_message(*err));
+            goto cleanup;
+        }
+        if ( (*err = grib_set_missing(outh, "jDirectionIncrement"))!=0 ){
+            fprintf(stderr,"GRIB_UTIL_SET_SPEC: Cannot set Dj to missing: %s\n",grib_get_error_message(*err));
+            goto cleanup;
+        }
+    }
+
+    /*grib_dump_content(outh, stdout,"debug", ~0, NULL);*/
+
     /* convert to second_order if not constant field */
     if (setSecondOrder ) {
         int ii=0;
@@ -1037,6 +1206,18 @@ grib_handle* grib_util_set_spec(grib_handle* h,
     if(packing_spec->editionNumber && packing_spec->editionNumber!=editionNumber)
         grib_set_long(outh,"edition", packing_spec->editionNumber);
 
+    if ( (*err = check_handle_against_spec(outh, editionNumber, spec)) != GRIB_SUCCESS)
+    {
+        grib_context* c=grib_context_get_default();
+        fprintf(stderr,"GRIB_UTIL_SET_SPEC: Geometry check failed! %s\n", grib_get_error_message(*err));
+        if (editionNumber == 1) {
+            fprintf(stderr,"Note: in GRIB edition 1 latitude and longitude values cannot be represented with sub-millidegree precision.\n");
+        }
+        if (c->write_on_fail)
+            grib_write_message(outh,"error.grib","w");
+        goto cleanup;
+    }
+
     if (h->context->debug==-1)
         printf("GRIB_API DEBUG: grib_util_set_spec end\n");
 
